@@ -7,7 +7,7 @@ test -z "$srcdir" && srcdir=.
 PKG_NAME="muffin"
 REQUIRED_AUTOMAKE_VERSION=1.10
 
-(test -f $srcdir/configure.in \
+(test -f $srcdir/configure.ac \
   && test -d $srcdir/src) || {
     echo -n "**Error**: Directory "\`$srcdir\'" does not look like the"
     echo " top-level muffin directory"
