@@ -29,7 +29,7 @@
 #include "workspace-private.h"
 #include "place.h"
 #include <meta/prefs.h>
-
+#include "xprops.h"
 #include <stdlib.h>
 #include <math.h>
 
@@ -204,15 +204,18 @@ static void place_window_if_needed       (MetaWindow     *window,
                                           ConstraintInfo *info);
 static void update_onscreen_requirements (MetaWindow     *window,
                                           ConstraintInfo *info);
-static void extend_by_frame              (MetaRectangle           *rect,
+static void extend_by_frame              (MetaWindow *window,
+                                          MetaRectangle           *rect,
                                           const MetaFrameBorders  *borders);
-static void unextend_by_frame            (MetaRectangle           *rect,
+static void unextend_by_frame            (MetaWindow *window,
+                                          MetaRectangle           *rect,
                                           const MetaFrameBorders  *borders);
-static inline void get_size_limits       (const MetaWindow        *window,
-                                          const MetaFrameBorders  *borders,
-                                          gboolean include_frame,
-                                          MetaRectangle *min_size,
-                                          MetaRectangle *max_size);
+
+static inline void get_size_limits   (const MetaWindow        *window,
+                                      const MetaFrameBorders  *borders,
+                                            gboolean           include_frame,
+                                            MetaRectangle     *min_size,
+                                            MetaRectangle     *max_size);
 
 typedef gboolean (* ConstraintFunc) (MetaWindow         *window,
                                      ConstraintInfo     *info,
@@ -446,13 +449,10 @@ setup_constraint_info (ConstraintInfo      *info,
                                          monitor_info->number);
 
   /* Workaround braindead legacy apps that don't know how to
-   * fullscreen themselves properly - don't get fooled by
-   * windows which hide their titlebar when maximized; that's
-   * not the same as fullscreen, even if there are no struts
-   * making the workarea smaller than the monitor.
+   * fullscreen themselves properly.
    */
   if (meta_prefs_get_force_fullscreen() &&
-      !window->hide_titlebar_when_maximized &&
+      (window->decorated && !meta_window_is_client_decorated (window)) &&
       meta_rectangle_equal (new, &monitor_info->rect) &&
       window->has_fullscreen_func &&
       !window->fullscreen)
@@ -571,6 +571,11 @@ place_window_if_needed(MetaWindow     *window,
                        .083 * info->work_area_monitor.height;
             }
 
+          /* idle_move_resize() uses the user_rect,
+           * so make sure it uses the placed coordinates.
+           */
+          window->user_rect = info->current;
+
           if (window->maximize_horizontally_after_placement ||
               window->maximize_vertically_after_placement)
             meta_window_maximize_internal (window,   
@@ -600,6 +605,39 @@ place_window_if_needed(MetaWindow     *window,
           meta_window_minimize (window);
           window->minimize_after_placement = FALSE;
         }
+      if (window->tile_after_placement)
+        {
+          window->tile_after_placement = FALSE;
+
+          gulong *tile_info = NULL;
+          int nitems;
+
+          if (meta_prop_get_cardinal_list (window->display,
+                                           window->xwindow,
+                                           window->display->atom__NET_WM_WINDOW_TILE_INFO,
+                                           &tile_info, &nitems))
+            {
+              if (nitems == 8)
+                {
+                  window->tile_mode = (MetaTileMode) tile_info[0];
+                  meta_window_move_resize_frame (window,
+                                                 TRUE,
+                                                 tile_info[2],
+                                                 tile_info[3],
+                                                 tile_info[4],
+                                                 tile_info[5]);
+                  window->custom_snap_size = tile_info[7] == 1;
+                  window->tile_monitor_number = tile_info[6];
+                  if (tile_info[1] == META_WINDOW_TILE_TYPE_SNAPPED)
+                    window->snap_queued = TRUE;
+                  else
+                    window->snap_queued = FALSE;
+                  meta_window_real_tile (window, TRUE);
+                  meta_XFree (tile_info);
+                }
+            }
+
+        }
     }
 }
 
@@ -642,7 +680,7 @@ update_onscreen_requirements (MetaWindow     *window,
   /* The require onscreen/on-single-monitor and titlebar_visible
    * stuff is relative to the outer window, not the inner
    */
-  extend_by_frame (&info->current, info->borders);
+  extend_by_frame (window, &info->current, info->borders);
 
   /* Update whether we want future constraint runs to require the
    * window to be on fully onscreen.
@@ -673,53 +711,80 @@ update_onscreen_requirements (MetaWindow     *window,
   /* Update whether we want future constraint runs to require the
    * titlebar to be visible.
    */
-  if (window->frame && window->decorated)
-    {
-      MetaRectangle titlebar_rect;
 
-      titlebar_rect = info->current;
-      titlebar_rect.height = info->borders->visible.top;
-      old = window->require_titlebar_visible;
-      window->require_titlebar_visible =
-        meta_rectangle_overlaps_with_region (info->usable_screen_region,
-                                             &titlebar_rect);
-      if (old ^ window->require_titlebar_visible)
-        meta_topic (META_DEBUG_GEOMETRY,
-                    "require_titlebar_visible for %s toggled to %s\n",
-                    window->desc,
-                    window->require_titlebar_visible ? "TRUE" : "FALSE");
-    }
+  MetaRectangle titlebar_rect;
+
+  meta_window_get_titlebar_rect (window, &titlebar_rect);
+
+  titlebar_rect.x += info->current.x;
+  titlebar_rect.y += info->current.y;
+
+  old = window->require_titlebar_visible;
+  window->require_titlebar_visible =
+    meta_rectangle_overlaps_with_region (info->usable_screen_region,
+                                         &titlebar_rect);
+
+  if (old ^ window->require_titlebar_visible)
+    meta_topic (META_DEBUG_GEOMETRY,
+                "require_titlebar_visible for %s toggled to %s\n",
+                window->desc,
+                window->require_titlebar_visible ? "TRUE" : "FALSE");
+
 
   /* Don't forget to restore the position of the window */
-  unextend_by_frame (&info->current, info->borders);
+  unextend_by_frame (window, &info->current, info->borders);
 }
 
 static void
-extend_by_frame (MetaRectangle           *rect,
-                 const MetaFrameBorders *borders)
+extend_by_frame (MetaWindow              *window,
+                 MetaRectangle           *rect,
+                 const MetaFrameBorders  *borders)
 {
-  rect->x -= borders->visible.left;
-  rect->y -= borders->visible.top;
-  rect->width  += borders->visible.left + borders->visible.right;
-  rect->height += borders->visible.top + borders->visible.bottom;
+  if (window->frame)
+    {
+      rect->x -= borders->visible.left;
+      rect->y -= borders->visible.top;
+      rect->width  += borders->visible.left + borders->visible.right;
+      rect->height += borders->visible.top + borders->visible.bottom;
+    }
+  else if (meta_window_is_client_decorated (window))
+    {
+      const GtkBorder *extents = &window->custom_frame_extents;
+      rect->x += extents->left;
+      rect->y += extents->top;
+      rect->width -= extents->left + extents->right;
+      rect->height -= extents->top + extents->bottom;
+    }
 }
 
 static void
-unextend_by_frame (MetaRectangle           *rect,
-                   const MetaFrameBorders *borders)
+unextend_by_frame (MetaWindow              *window,
+                   MetaRectangle           *rect,
+                   const MetaFrameBorders  *borders)
 {
-  rect->x += borders->visible.left;
-  rect->y += borders->visible.top;
-  rect->width  -= borders->visible.left + borders->visible.right;
-  rect->height -= borders->visible.top + borders->visible.bottom;
+  if (window->frame)
+    {
+      rect->x += borders->visible.left;
+      rect->y += borders->visible.top;
+      rect->width  -= borders->visible.left + borders->visible.right;
+      rect->height -= borders->visible.top + borders->visible.bottom;
+    }
+  else if (meta_window_is_client_decorated (window))
+    {
+      const GtkBorder *extents = &window->custom_frame_extents;
+      rect->x -= extents->left;
+      rect->y -= extents->top;
+      rect->width += extents->left + extents->right;
+      rect->height += extents->top + extents->bottom;
+    }
 }
 
 static inline void
-get_size_limits (const MetaWindow        *window,
+get_size_limits (const MetaWindow       *window,
                  const MetaFrameBorders *borders,
-                 gboolean                 include_frame,
-                 MetaRectangle *min_size,
-                 MetaRectangle *max_size)
+                       gboolean          include_frame,
+                       MetaRectangle    *min_size,
+                       MetaRectangle    *max_size)
 {
   /* We pack the results into MetaRectangle structs just for convienience; we
    * don't actually use the position of those rects.
@@ -764,21 +829,12 @@ constrain_modal_dialog (MetaWindow         *window,
     return TRUE;
 
   x = parent->rect.x + (parent->rect.width / 2  - info->current.width / 2);
-  y = 0;
+  y = parent->rect.y + (parent->rect.height / 2 - info->current.height / 2);
   if (parent->frame)
     {
-      MetaFrameBorders borders;
-
       x += parent->frame->rect.x;
       y += parent->frame->rect.y;
-
-      meta_frame_calc_borders (parent->frame, &borders);
-      y += borders.total.top;
-
-      y += info->borders->visible.top;
     }
-  else
-    y = parent->rect.y + info->borders->visible.top;
 
   constraint_already_satisfied = (x == info->current.x) && (y == info->current.y);
 
@@ -812,15 +868,11 @@ constrain_maximization (MetaWindow         *window,
 
   /* Determine whether constraint applies; exit if it doesn't */
   if ((!window->maximized_horizontally && !window->maximized_vertically) ||
-      META_WINDOW_TILED_SIDE_BY_SIDE (window))
+      META_WINDOW_TILED_OR_SNAPPED (window))
     return TRUE;
 
   /* Calculate target_size = maximized size of (window + frame) */
-  if (META_WINDOW_TILED_MAXIMIZED (window))
-    {
-      meta_window_get_current_tile_area (window, &target_size);
-    }
-  else if (META_WINDOW_MAXIMIZED (window))
+  if (META_WINDOW_MAXIMIZED (window) && g_list_length (window->workspace->snapped_windows) == 0)
     {
       target_size = info->work_area_monitor;
     }
@@ -842,15 +894,45 @@ constrain_maximization (MetaWindow         *window,
         direction = META_DIRECTION_VERTICAL;
       active_workspace_struts = window->screen->active_workspace->all_struts;
 
-      target_size = info->current;
-      extend_by_frame (&target_size, info->borders);
-      meta_rectangle_expand_to_avoiding_struts (&target_size,
-                                                &info->entire_monitor,
-                                                direction,
-                                                active_workspace_struts);
+      if (g_list_length (window->screen->active_workspace->snapped_windows) > 0) {
+        GList *tmp = window->screen->active_workspace->snapped_windows;
+        GSList *snapped_windows_as_struts = NULL;
+        while (tmp) {
+            if (tmp->data == window || META_WINDOW (tmp->data)->minimized ||
+                meta_window_get_monitor (window) != meta_window_get_monitor (META_WINDOW (tmp->data))) {
+                tmp = tmp->next;
+                continue;
+            }
+            MetaStrut *strut = g_slice_new0 (MetaStrut);
+            MetaSide side;
+            MetaRectangle rect;
+            meta_window_get_outer_rect (META_WINDOW (tmp->data), &rect);
+            side = meta_window_get_tile_side (META_WINDOW (tmp->data));
+            strut->rect = rect;
+            strut->side = side;
+            snapped_windows_as_struts = g_slist_prepend (snapped_windows_as_struts, strut);
+            tmp = tmp->next;
+        }
+
+        target_size = info->current;
+        extend_by_frame (window, &target_size, info->borders);
+        meta_rectangle_expand_to_snapped_borders (&target_size,
+                                                  &info->entire_monitor,
+                                                   active_workspace_struts,
+                                                   snapped_windows_as_struts,
+                                                   &window->user_rect);
+        g_slist_free (snapped_windows_as_struts);
+      } else {
+          target_size = info->current;
+          extend_by_frame (window, &target_size, info->borders);
+          meta_rectangle_expand_to_avoiding_struts (&target_size,
+                                                    &info->entire_monitor,
+                                                    direction,
+                                                    active_workspace_struts);
+      }
    }
   /* Now make target_size = maximized size of client window */
-  unextend_by_frame (&target_size, info->borders);
+  unextend_by_frame (window, &target_size, info->borders);
 
   /* Check min size constraints; max size constraints are ignored for maximized
    * windows, as per bug 327543.
@@ -894,6 +976,7 @@ constrain_tiling (MetaWindow         *window,
 {
   MetaRectangle target_size;
   MetaRectangle min_size, max_size;
+  MetaRectangle actual_position;
   gboolean hminbad, vminbad;
   gboolean horiz_equal, vert_equal;
   gboolean constraint_already_satisfied;
@@ -902,14 +985,109 @@ constrain_tiling (MetaWindow         *window,
     return TRUE;
 
   /* Determine whether constraint applies; exit if it doesn't */
-  if (!META_WINDOW_TILED_SIDE_BY_SIDE (window))
+  if (!META_WINDOW_TILED_OR_SNAPPED (window))
     return TRUE;
 
   /* Calculate target_size - as the tile previews need this as well, we
    * use an external function for the actual calculation
    */
-  meta_window_get_current_tile_area (window, &target_size);
-  unextend_by_frame (&target_size, info->borders);
+  if (window->tile_mode != META_TILE_NONE)
+    meta_window_get_current_tile_area (window, &target_size);
+  else
+    return TRUE;
+
+  meta_window_get_outer_rect (window, &actual_position);
+
+  if (window->custom_snap_size) {
+      switch (window->tile_mode) {
+        case META_TILE_LEFT:
+            target_size.width = BOX_RIGHT (actual_position) - target_size.x;
+            break;
+        case META_TILE_RIGHT:
+            target_size.width = BOX_RIGHT (target_size) - BOX_LEFT (actual_position);
+            target_size.x = BOX_LEFT (actual_position);
+            break;
+        case META_TILE_TOP:
+            target_size.height = BOX_BOTTOM (actual_position) - BOX_TOP (target_size);
+            break;
+        case META_TILE_BOTTOM:
+            target_size.height = BOX_BOTTOM (target_size) - BOX_TOP (actual_position);
+            target_size.y = BOX_TOP (actual_position);
+            break;
+        case META_TILE_ULC:
+            if (GRAB_OP (window) == META_GRAB_OP_RESIZING_S ||
+                GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_S) {
+                target_size.width = window->snapped_rect.width;
+                target_size.height = BOX_BOTTOM (actual_position) - BOX_TOP (target_size);
+            } else if (GRAB_OP (window) == META_GRAB_OP_RESIZING_E ||
+                       GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_E) {
+                target_size.width = BOX_RIGHT (actual_position) - target_size.x;
+                target_size.height = window->snapped_rect.height;
+            } else {
+                target_size.width = BOX_RIGHT (actual_position) - target_size.x;
+                target_size.height = BOX_BOTTOM (actual_position) - BOX_TOP (target_size);
+            }
+            break;
+        case META_TILE_LLC:
+            if (GRAB_OP (window) == META_GRAB_OP_RESIZING_N ||
+                GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_N) {
+                target_size.width = window->snapped_rect.width;
+                target_size.height = BOX_BOTTOM (target_size) - BOX_TOP (actual_position);
+                target_size.y = BOX_TOP (actual_position);
+            } else if (GRAB_OP (window) == META_GRAB_OP_RESIZING_E ||
+                       GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_E) {
+                target_size.width = BOX_RIGHT (actual_position) - target_size.x;
+                target_size.height = window->snapped_rect.height;
+                target_size.y = window->snapped_rect.y;
+            } else {
+                target_size.width = BOX_RIGHT (actual_position) - target_size.x;
+                target_size.height = BOX_BOTTOM (target_size) - BOX_TOP (actual_position);
+                target_size.y = BOX_TOP (actual_position);
+            }
+            break;
+        case META_TILE_URC:
+            if (GRAB_OP (window) == META_GRAB_OP_RESIZING_W ||
+                GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_W) {
+                target_size.width = BOX_RIGHT (target_size) - BOX_LEFT (actual_position);
+                target_size.x = BOX_LEFT (actual_position);
+                target_size.height = window->snapped_rect.height;
+            } else if (GRAB_OP (window) == META_GRAB_OP_RESIZING_S ||
+                       GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_S) {
+                target_size.width = window->snapped_rect.width;
+                target_size.x = window->snapped_rect.x;
+                target_size.height = BOX_BOTTOM (actual_position) - BOX_TOP (target_size);
+            } else {
+                target_size.width = BOX_RIGHT (target_size) - BOX_LEFT (actual_position);
+                target_size.x = BOX_LEFT (actual_position);
+                target_size.height = BOX_BOTTOM (actual_position) - BOX_TOP (target_size);
+            }
+            break;
+        case META_TILE_LRC:
+            if (GRAB_OP (window) == META_GRAB_OP_RESIZING_W ||
+                GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_W) {
+                target_size.width = BOX_RIGHT (target_size) - BOX_LEFT (actual_position);
+                target_size.x = BOX_LEFT (actual_position);
+                target_size.height = window->snapped_rect.height;
+                target_size.y = window->snapped_rect.y;
+            } else if (GRAB_OP (window) == META_GRAB_OP_RESIZING_N ||
+                       GRAB_OP (window) == META_GRAB_OP_KEYBOARD_RESIZING_N) {
+                target_size.width = window->snapped_rect.width;
+                target_size.x = window->snapped_rect.x;
+                target_size.height = BOX_BOTTOM (target_size) - BOX_TOP (actual_position);
+                target_size.y = BOX_TOP (actual_position);
+            } else {
+                target_size.width = BOX_RIGHT (target_size) - BOX_LEFT (actual_position);
+                target_size.x = BOX_LEFT (actual_position);
+                target_size.height = BOX_BOTTOM (target_size) - BOX_TOP (actual_position);
+                target_size.y = BOX_TOP (actual_position);
+            }
+            break;
+        default:
+            break;
+      }
+  }
+
+  unextend_by_frame (window, &target_size, info->borders);
 
   /* Check min size constraints; max size constraints are ignored as for
    * maximized windows.
@@ -990,8 +1168,9 @@ constrain_size_increments (MetaWindow         *window,
 
   /* Determine whether constraint applies; exit if it doesn't */
   if (META_WINDOW_MAXIMIZED (window) || window->fullscreen || 
-      META_WINDOW_TILED_SIDE_BY_SIDE (window) ||
-      info->action_type == ACTION_MOVE)
+      META_WINDOW_TILED_OR_SNAPPED (window) ||
+      info->action_type == ACTION_MOVE ||
+      window->resizing_tile_type != META_WINDOW_TILE_TYPE_NONE)
     return TRUE;
 
   /* Determine whether constraint is already satisfied; exit if it is */
@@ -1122,7 +1301,7 @@ constrain_aspect_ratio (MetaWindow         *window,
   constraints_are_inconsistent = minr > maxr;
   if (constraints_are_inconsistent ||
       META_WINDOW_MAXIMIZED (window) || window->fullscreen || 
-      META_WINDOW_TILED_SIDE_BY_SIDE (window) ||
+      META_WINDOW_TILED_OR_SNAPPED (window) ||
       info->action_type == ACTION_MOVE)
     return TRUE;
 
@@ -1259,7 +1438,7 @@ do_screen_and_monitor_relative_constraints (
   /* Determine whether constraint applies; exit if it doesn't */
   how_far_it_can_be_smushed = info->current;
   get_size_limits (window, info->borders, TRUE, &min_size, &max_size);
-  extend_by_frame (&info->current, info->borders);
+  extend_by_frame (window, &info->current, info->borders);
 
   if (info->action_type != ACTION_MOVE)
     {
@@ -1279,7 +1458,7 @@ do_screen_and_monitor_relative_constraints (
                                         &info->current);
   if (exit_early || constraint_satisfied || check_only)
     {
-      unextend_by_frame (&info->current, info->borders);
+      unextend_by_frame (window, &info->current, info->borders);
       return constraint_satisfied;
     }
 
@@ -1298,12 +1477,14 @@ do_screen_and_monitor_relative_constraints (
                                    info->fixed_directions,
                                    &info->current);
   else
-    /* For everything else, shove the rectangle into the relevant region */
-    meta_rectangle_shove_into_region (region_spanning_rectangles,
-                                      info->fixed_directions,
-                                      &info->current);
+    {
+      /* For everything else, shove the rectangle into the relevant region */
+      meta_rectangle_shove_into_region (region_spanning_rectangles,
+                                        info->fixed_directions,
+                                        &info->current);
+    }
 
-  unextend_by_frame (&info->current, info->borders);
+  unextend_by_frame (window, &info->current, info->borders);
   return TRUE;
 }
 
@@ -1325,7 +1506,7 @@ constrain_to_single_monitor (MetaWindow         *window,
       window->type == META_WINDOW_DOCK      ||
       window->screen->n_monitor_infos == 1  ||
       !window->require_on_single_monitor    ||
-      !window->frame                        ||
+      (!window->frame && !meta_window_is_client_decorated (window)) ||
       info->is_user_action)
     return TRUE;
 
@@ -1374,6 +1555,7 @@ constrain_titlebar_visible (MetaWindow         *window,
   int bottom_amount;
   int horiz_amount_offscreen, vert_amount_offscreen;
   int horiz_amount_onscreen,  vert_amount_onscreen;
+  int scale = meta_prefs_get_ui_scale ();
 
   if (priority > PRIORITY_TITLEBAR_VISIBLE)
     return TRUE;
@@ -1392,7 +1574,6 @@ constrain_titlebar_visible (MetaWindow         *window,
       window->type == META_WINDOW_DOCK    ||
       window->fullscreen                  ||
       !window->require_titlebar_visible   ||
-      !window->decorated                  ||
       unconstrained_user_action)
     return TRUE;
 
@@ -1405,8 +1586,8 @@ constrain_titlebar_visible (MetaWindow         *window,
    */
   horiz_amount_onscreen = info->current.width  / 4;
   vert_amount_onscreen  = info->current.height / 4;
-  horiz_amount_onscreen = CLAMP (horiz_amount_onscreen, 10, 75);
-  vert_amount_onscreen  = CLAMP (vert_amount_onscreen,  10, 75);
+  horiz_amount_onscreen = CLAMP (horiz_amount_onscreen, 10 * scale, 75 * scale);
+  vert_amount_onscreen  = CLAMP (vert_amount_onscreen,  10 * scale, 75 * scale);
   horiz_amount_offscreen = info->current.width - horiz_amount_onscreen;
   vert_amount_offscreen  = info->current.height - vert_amount_onscreen;
   horiz_amount_offscreen = MAX (horiz_amount_offscreen, 0);
@@ -1419,6 +1600,11 @@ constrain_titlebar_visible (MetaWindow         *window,
       bottom_amount = info->current.height + info->borders->visible.bottom;
       vert_amount_onscreen = info->borders->visible.top;
     }
+  else if (meta_window_is_client_decorated (window))
+    {
+      vert_amount_onscreen = CSD_TITLEBAR_HEIGHT * scale; /* Hardcoded for now, we don't get this from Gtk */
+      bottom_amount = vert_amount_offscreen = MAX ((info->current.height - (vert_amount_onscreen * 2)), 0);
+    }
   else
     bottom_amount = vert_amount_offscreen;
 
@@ -1444,7 +1630,6 @@ constrain_titlebar_visible (MetaWindow         *window,
                                               -bottom_amount,
                                               horiz_amount_onscreen,
                                               vert_amount_onscreen);
-
   return retval;
 }
 
@@ -1458,6 +1643,7 @@ constrain_partially_onscreen (MetaWindow         *window,
   int top_amount, bottom_amount;
   int horiz_amount_offscreen, vert_amount_offscreen;
   int horiz_amount_onscreen,  vert_amount_onscreen;
+  int scale = meta_prefs_get_ui_scale ();
 
   if (priority > PRIORITY_PARTIALLY_VISIBLE_ON_WORKAREA)
     return TRUE;
@@ -1479,8 +1665,8 @@ constrain_partially_onscreen (MetaWindow         *window,
    */
   horiz_amount_onscreen = info->current.width  / 4;
   vert_amount_onscreen  = info->current.height / 4;
-  horiz_amount_onscreen = CLAMP (horiz_amount_onscreen, 10, 75);
-  vert_amount_onscreen  = CLAMP (vert_amount_onscreen,  10, 75);
+  horiz_amount_onscreen = CLAMP (horiz_amount_onscreen, 10 * scale, 75 * scale);
+  vert_amount_onscreen  = CLAMP (vert_amount_onscreen,  10 * scale, 75 * scale);
   horiz_amount_offscreen = info->current.width - horiz_amount_onscreen;
   vert_amount_offscreen  = info->current.height - vert_amount_onscreen;
   horiz_amount_offscreen = MAX (horiz_amount_offscreen, 0);
@@ -1494,6 +1680,11 @@ constrain_partially_onscreen (MetaWindow         *window,
       bottom_amount = info->current.height + info->borders->visible.bottom;
       vert_amount_onscreen = info->borders->visible.top;
     }
+  else if (meta_window_is_client_decorated (window))
+    {
+      top_amount = vert_amount_onscreen = CSD_TITLEBAR_HEIGHT * scale; /* Hardcoded for now, we don't get this from Gtk */
+      bottom_amount = vert_amount_offscreen = MAX ((info->current.height - (vert_amount_onscreen * 2)), 0);
+    }
   else
     bottom_amount = vert_amount_offscreen;
 
