@@ -24,6 +24,12 @@
  * 02110-1335, USA.
  */
 
+/**
+ * SECTION:prefs
+ * @title: Preferences
+ * @short_description: Muffin preferences
+ */
+
 #include <config.h>
 #include <meta/prefs.h>
 #include "ui.h"
@@ -44,7 +50,7 @@
  */
 #define KEY_TITLEBAR_FONT "titlebar-font"
 #define KEY_NUM_WORKSPACES "num-workspaces"
-#define KEY_WORKSPACE_NAMES "workspace-name-overrides"
+#define KEY_WORKSPACE_NAMES "workspace-names"
 #define KEY_WORKSPACE_CYCLE "workspace-cycle"
 
 /* Keys from "foreign" schemas */
@@ -55,17 +61,17 @@
 
 #define KEY_MIN_WINDOW_OPACITY "min-window-opacity"
 #define KEY_WS_NAMES_GNOME "workspace-names"
-#define KEY_OVERLAY_KEY "overlay-key"
 #define KEY_LIVE_HIDDEN_WINDOWS "live-hidden-windows"
 #define KEY_WORKSPACES_ONLY_ON_PRIMARY "workspaces-only-on-primary"
-#define KEY_NO_TAB_POPUP "no-tab-popup"
+
+#define KEY_MOUSEWHEEL_ZOOM_ENABLED "screen-magnifier-enabled"
 
 /* These are the different schemas we are keeping
  * a GSettings instance for */
 #define SCHEMA_GENERAL         "org.cinnamon.desktop.wm.preferences"
-#define SCHEMA_CINNAMON          "org.cinnamon"
 #define SCHEMA_MUFFIN          "org.cinnamon.muffin"
 #define SCHEMA_INTERFACE       "org.cinnamon.desktop.interface"
+#define SCHEMA_A11Y_APPLICATIONS "org.cinnamon.desktop.a11y.applications"
 
 #define SETTINGS(s) g_hash_table_lookup (settings_schemas, (s))
 
@@ -77,6 +83,8 @@ static GHashTable *settings_schemas;
 static gboolean use_system_font = FALSE;
 static PangoFontDescription *titlebar_font = NULL;
 static MetaVirtualModifier mouse_button_mods = Mod1Mask;
+static MetaVirtualModifier mouse_button_zoom_mods = Mod1Mask;
+static gboolean mouse_zoom_enabled = FALSE;
 static CDesktopFocusMode focus_mode = C_DESKTOP_FOCUS_MODE_CLICK;
 static CDesktopFocusNewWindows focus_new_windows = C_DESKTOP_FOCUS_NEW_WINDOWS_SMART;
 static gboolean raise_on_click = TRUE;
@@ -89,6 +97,7 @@ static CDesktopTitlebarAction action_middle_click_titlebar = C_DESKTOP_TITLEBAR_
 static CDesktopTitlebarAction action_right_click_titlebar = C_DESKTOP_TITLEBAR_ACTION_MENU;
 static CDesktopTitlebarScrollAction action_scroll_titlebar = C_DESKTOP_TITLEBAR_SCROLL_ACTION_NONE;
 static gboolean dynamic_workspaces = FALSE;
+static gboolean unredirect_fullscreen_windows = FALSE;
 static gboolean application_based = FALSE;
 static gboolean disable_workarounds = FALSE;
 static gboolean auto_raise = FALSE;
@@ -118,7 +127,6 @@ static char **workspace_names = NULL;
 static gboolean live_hidden_windows = FALSE;
 static gboolean workspaces_only_on_primary = FALSE;
 
-static gboolean no_tab_popup = FALSE;
 static gboolean legacy_snap = FALSE;
 static gboolean invert_workspace_flip = FALSE;
 static gboolean tile_maximize = FALSE;
@@ -146,6 +154,7 @@ static void maybe_give_disable_workarounds_warning (void);
 static gboolean titlebar_handler (GVariant*, gpointer*, gpointer);
 static gboolean theme_name_handler (GVariant*, gpointer*, gpointer);
 static gboolean mouse_button_mods_handler (GVariant*, gpointer*, gpointer);
+static gboolean mouse_button_zoom_mods_handler (GVariant*, gpointer*, gpointer);
 static gboolean snap_modifier_handler (GVariant*, gpointer*, gpointer);
 static gboolean button_layout_handler (GVariant*, gpointer*, gpointer);
 
@@ -326,6 +335,13 @@ static MetaBoolPreference preferences_bool[] =
       },
       &dynamic_workspaces,
     },
+    {
+      { "unredirect-fullscreen-windows",
+        SCHEMA_MUFFIN,
+        META_PREF_UNREDIRECT_FULLSCREEN_WINDOWS,
+      },
+      &unredirect_fullscreen_windows,
+    },
     {
       { "application-based",
         SCHEMA_GENERAL,
@@ -368,6 +384,13 @@ static MetaBoolPreference preferences_bool[] =
       },
       &gnome_accessibility,
     },
+    {
+      { KEY_MOUSEWHEEL_ZOOM_ENABLED,
+        SCHEMA_A11Y_APPLICATIONS,
+        META_PREF_MOUSE_ZOOM_ENABLED,
+      },
+      &mouse_zoom_enabled,
+    },
     {
       { KEY_GNOME_ANIMATIONS,
         SCHEMA_INTERFACE,
@@ -403,13 +426,6 @@ static MetaBoolPreference preferences_bool[] =
       },
       &workspaces_only_on_primary,
     },
-    {
-      { KEY_NO_TAB_POPUP,
-        SCHEMA_MUFFIN,
-        META_PREF_NO_TAB_POPUP,
-      },
-      &no_tab_popup,
-    },
     {
       { "legacy-snap",
         SCHEMA_MUFFIN,
@@ -444,6 +460,14 @@ static MetaStringPreference preferences_string[] =
       mouse_button_mods_handler,
       NULL,
     },
+    {
+      { "mouse-button-zoom-modifier",
+        SCHEMA_GENERAL,
+        META_PREF_MOUSE_BUTTON_ZOOM_MODS,
+      },
+      mouse_button_zoom_mods_handler,
+      NULL,
+    },
     {
       { "theme",
         SCHEMA_GENERAL,
@@ -918,10 +942,6 @@ meta_prefs_init (void)
   g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
   g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_MUFFIN), settings);
 
-  settings = g_settings_new (SCHEMA_CINNAMON);
-  g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
-  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_CINNAMON), settings);
-
   /* Individual keys we watch outside of our schemas */
   settings = g_settings_new (SCHEMA_INTERFACE);
   g_signal_connect (settings, "changed::" KEY_GNOME_ACCESSIBILITY,
@@ -934,6 +954,10 @@ meta_prefs_init (void)
                     G_CALLBACK (settings_changed), NULL);
   g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_INTERFACE), settings);
 
+  settings = g_settings_new (SCHEMA_A11Y_APPLICATIONS);
+  g_signal_connect (settings, "changed::" KEY_MOUSEWHEEL_ZOOM_ENABLED,
+                    G_CALLBACK (settings_changed), NULL);
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_A11Y_APPLICATIONS), settings);
 
   for (tmp = overridden_keys; tmp; tmp = tmp->next)
     {
@@ -1125,9 +1149,6 @@ settings_changed (GSettings *settings,
       return;
     }
 
-  if (strcmp(schema, SCHEMA_CINNAMON) == 0)
-    return;
-
   value = g_settings_get_value (settings, key);
   type = g_variant_get_type (value);
 
@@ -1154,10 +1175,6 @@ settings_changed (GSettings *settings,
       else
         handle_preference_update_string (settings, key);
     }
-  else if (g_str_equal (key, KEY_OVERLAY_KEY))
-    {
-      queue_changed (META_PREF_KEYBINDINGS);
-    }
   else if (g_str_equal (key, KEY_WS_NAMES_GNOME))
     {
       return;
@@ -1209,6 +1226,18 @@ meta_prefs_get_mouse_button_mods  (void)
   return mouse_button_mods;
 }
 
+MetaVirtualModifier
+meta_prefs_get_mouse_button_zoom_mods  (void)
+{
+  return mouse_button_zoom_mods;
+}
+
+gboolean
+meta_prefs_get_mouse_zoom_enabled (void)
+{
+  return mouse_zoom_enabled;
+}
+
 CDesktopFocusMode
 meta_prefs_get_focus_mode (void)
 {
@@ -1359,6 +1388,42 @@ mouse_button_mods_handler (GVariant *value,
   return TRUE;
 }
 
+static gboolean
+mouse_button_zoom_mods_handler (GVariant *value,
+                                gpointer *result,
+                                gpointer  data)
+{
+  MetaVirtualModifier mods;
+  const gchar *string_value;
+
+  *result = NULL; /* ignored */
+  string_value = g_variant_get_string (value, NULL);
+
+  if (!string_value || !meta_ui_parse_modifier (string_value, &mods))
+    {
+      meta_topic (META_DEBUG_KEYBINDINGS,
+                  "Failed to parse new GSettings value\n");
+
+      meta_warning (_("\"%s\" found in configuration database is "
+                      "not a valid value for mouse button zoom modifier\n"),
+                    string_value);
+
+      return FALSE;
+    }
+
+  meta_topic (META_DEBUG_KEYBINDINGS,
+              "Mouse zoom modifier has new GSettings value \"%s\"\n",
+              string_value);
+
+  if (mods != mouse_button_zoom_mods)
+    {
+      mouse_button_zoom_mods = mods;
+      queue_changed (META_PREF_MOUSE_BUTTON_ZOOM_MODS);
+    }
+
+  return TRUE;
+}
+
 static gboolean
 snap_modifier_handler (GVariant *value,
                        gpointer *result,
@@ -1680,6 +1745,12 @@ meta_prefs_get_dynamic_workspaces (void)
   return dynamic_workspaces;
 }
 
+gboolean
+meta_prefs_get_unredirect_fullscreen_windows (void)
+{
+  return unredirect_fullscreen_windows;
+}
+
 gboolean
 meta_prefs_get_application_based (void)
 {
@@ -1704,6 +1775,12 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_MOUSE_BUTTON_MODS:
       return "MOUSE_BUTTON_MODS";
 
+    case META_PREF_MOUSE_BUTTON_ZOOM_MODS:
+      return "MOUSE_BUTTON_ZOOM_MODS";
+
+    case META_PREF_MOUSE_ZOOM_ENABLED:
+      return "MOUSE_ZOOM_ENABLED";
+
     case META_PREF_FOCUS_MODE:
       return "FOCUS_MODE";
 
@@ -1797,9 +1874,6 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_WORKSPACE_CYCLE:
       return "WORKSPACE_CYCLE";
 
-    case META_PREF_NO_TAB_POPUP:
-      return "NO_TAB_POPUP";
-
     case META_PREF_DRAGGABLE_BORDER_WIDTH:
       return "DRAGGABLE_BORDER_WIDTH";
 
@@ -1812,6 +1886,9 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_DYNAMIC_WORKSPACES:
       return "DYNAMIC_WORKSPACES";
 
+    case META_PREF_UNREDIRECT_FULLSCREEN_WINDOWS:
+      return "UNREDIRECT_FULLSCREEN_WINDOWS";
+
     case META_PREF_SNAP_MODIFIER:
       return "SNAP_MODIFIER";
 
@@ -1851,8 +1928,6 @@ meta_prefs_set_num_workspaces (int n_workspaces)
 
 static GHashTable *key_bindings;
 
-static MetaKeyCombo overlay_key_combo = { 0, 0, 0 };
-
 static void
 meta_key_pref_free (MetaKeyPref *pref)
 {
@@ -1864,37 +1939,11 @@ meta_key_pref_free (MetaKeyPref *pref)
   g_free (pref);
 }
 
-/* These bindings are for modifiers alone, so they need special handling */
-static void
-init_special_bindings (void)
-{
-  char *val;
-  
-  /* Default values for bindings which are global, but take special handling */
-  meta_ui_parse_accelerator ("Super_L", &overlay_key_combo.keysym, 
-                             &overlay_key_combo.keycode, 
-                             &overlay_key_combo.modifiers);
-
-  val = g_settings_get_string (SETTINGS (SCHEMA_MUFFIN), KEY_OVERLAY_KEY);
-    
-  if (val && meta_ui_parse_accelerator (val, &overlay_key_combo.keysym, 
-                                        &overlay_key_combo.keycode, 
-                                        &overlay_key_combo.modifiers))
-    ;
-  else
-    {
-      meta_topic (META_DEBUG_KEYBINDINGS,
-                  "Failed to parse value for overlay_key\n");
-    }
-  g_free (val);
-}
-
 static void
 init_bindings (void)
 {
   key_bindings = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
                                         (GDestroyNotify)meta_key_pref_free);
-  init_special_bindings ();  
 }
 
 static void
@@ -1994,7 +2043,7 @@ update_workspace_names (void)
   int n_workspace_names, n_names;
   gboolean changed = FALSE;
 
-  names = g_settings_get_strv (SETTINGS (SCHEMA_CINNAMON), KEY_WORKSPACE_NAMES);
+  names = g_settings_get_strv (SETTINGS (SCHEMA_GENERAL), KEY_WORKSPACE_NAMES);
   n_names = g_strv_length (names);
   n_workspace_names = workspace_names ? g_strv_length (workspace_names) : 0;
 
@@ -2099,7 +2148,7 @@ meta_prefs_change_workspace_name (int         num,
       g_variant_builder_add (&builder, "s", value);
     }
 
-  g_settings_set_value (SETTINGS (SCHEMA_CINNAMON), KEY_WORKSPACE_NAMES,
+  g_settings_set_value (SETTINGS (SCHEMA_GENERAL), KEY_WORKSPACE_NAMES,
                         g_variant_builder_end (&builder));
 }
 
@@ -2217,9 +2266,9 @@ meta_prefs_remove_keybinding (const char *name)
 
 LOCAL_SYMBOL gboolean
 meta_prefs_add_custom_keybinding (const char           *name,
-                                const char           *binding,
-                                MetaKeyBindingAction  action,
-                                MetaKeyBindingFlags   flags)
+                                  const char          **bindings,
+                                  MetaKeyBindingAction  action,
+                                  MetaKeyBindingFlags   flags)
 {
   MetaKeyPref  *pref;
 
@@ -2232,16 +2281,14 @@ meta_prefs_add_custom_keybinding (const char           *name,
 
   pref = g_new0 (MetaKeyPref, 1);
   pref->name = g_strdup (name);
-  pref->schema = g_strdup (binding);
+  pref->schema = NULL;
   pref->action = action;
   pref->bindings = NULL;
   pref->add_shift = (flags & META_KEY_BINDING_REVERSES) != 0;
   pref->per_window = (flags & META_KEY_BINDING_PER_WINDOW) != 0;
   pref->builtin = (flags & META_KEY_BINDING_BUILTIN) != 0;
-  
-  char **strokes = g_strsplit(binding, "XYZZY", 1);
-  update_binding (pref, strokes);
-  g_strfreev (strokes);
+
+  update_binding (pref, (gchar **)bindings);
 
   g_hash_table_insert (key_bindings, g_strdup (name), pref);
 
@@ -2278,12 +2325,6 @@ meta_prefs_get_keybindings ()
   return g_hash_table_get_values (key_bindings);
 }
 
-void 
-meta_prefs_get_overlay_binding (MetaKeyCombo *combo)
-{
-  *combo = overlay_key_combo;
-}
-
 CDesktopTitlebarAction
 meta_prefs_get_action_double_click_titlebar (void)
 {
@@ -2429,28 +2470,12 @@ meta_prefs_get_workspaces_only_on_primary (void)
   return workspaces_only_on_primary;
 }
 
-gboolean
-meta_prefs_get_no_tab_popup (void)
-{
-  return no_tab_popup;
-}
-
 gboolean
 meta_prefs_get_legacy_snap (void)
 {
   return legacy_snap;
 }
 
-void
-meta_prefs_set_no_tab_popup (gboolean whether)
-{
-  MetaBasePreference *pref = NULL;
-
-  find_pref (preferences_bool, sizeof(MetaBoolPreference),
-             KEY_NO_TAB_POPUP, &pref);
-  g_settings_set_boolean (SETTINGS (pref->schema), KEY_NO_TAB_POPUP, whether);
-}
-
 int
 meta_prefs_get_draggable_border_width (void)
 {
@@ -2504,3 +2529,9 @@ meta_prefs_get_min_win_opacity (void)
 {
   return min_window_opacity;
 }
+
+gint
+meta_prefs_get_ui_scale (void)
+{
+  return ui_scale;
+}
