@@ -1,7 +1,5 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 
-/* Muffin Workspaces */
-
 /* 
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2003 Rob Adams
@@ -23,6 +21,19 @@
  * 02110-1335, USA.
  */
 
+/**
+ * SECTION:workspace
+ * @title: MetaWorkspace
+ * @short_description: Workspaces
+ *
+ * A workspace is a set of windows which all live on the same
+ * screen.  (You may also see the name "desktop" around the place,
+ * which is the EWMH's name for the same thing.)  Only one workspace
+ * of a screen may be active at once; all windows on all other workspaces
+ * are unmapped.
+ */
+
+
 #include <config.h>
 #include "screen-private.h"
 #include <meta/workspace.h>
@@ -68,6 +79,8 @@ static void
 meta_workspace_finalize (GObject *object)
 {
   /* Actual freeing done in meta_workspace_remove() for now */
+  
+  G_OBJECT_CLASS (meta_workspace_parent_class)->finalize (object);
 }
 
 static void
@@ -183,7 +196,7 @@ meta_workspace_new (MetaScreen *screen)
   workspace->screen_edges = NULL;
   workspace->monitor_edges = NULL;
   workspace->list_containing_self = g_list_prepend (NULL, workspace);
-
+  workspace->snapped_windows = NULL;
   workspace->builtin_struts = NULL;
   workspace->all_struts = NULL;
 
@@ -192,7 +205,7 @@ meta_workspace_new (MetaScreen *screen)
   return workspace;
 }
 
-/** Foreach function for workspace_free_struts() */
+/* Foreach function for workspace_free_struts() */
 static void
 free_this (gpointer candidate, gpointer dummy)
 {
@@ -992,18 +1005,6 @@ ensure_work_areas_validated (MetaWorkspace *workspace)
 
   /* We're all done, YAAY!  Record that everything has been validated. */
   workspace->work_areas_invalid = FALSE;
-
-  {
-    /*
-     * Notify the compositor that the workspace geometry has changed.
-     */
-    MetaScreen     *screen = workspace->screen;
-    MetaDisplay    *display = meta_screen_get_display (screen);
-    MetaCompositor *comp = meta_display_get_compositor (display);
-
-    if (comp)
-      meta_compositor_update_workspace_geometry (comp, workspace);
-  }
 }
 
 static gboolean
@@ -1036,6 +1037,45 @@ void
 meta_workspace_set_builtin_struts (MetaWorkspace *workspace,
                                    GSList        *struts)
 {
+  MetaScreen *screen = workspace->screen;
+  GSList *l;
+
+  for (l = struts; l; l = l->next)
+    {
+      MetaStrut *strut = l->data;
+      int idx = meta_screen_get_monitor_index_for_rect (screen, &strut->rect);
+
+      switch (strut->side)
+        {
+        case META_SIDE_TOP:
+          if (meta_screen_get_monitor_neighbor (screen, idx, META_SCREEN_UP))
+            continue;
+
+          strut->rect.height += strut->rect.y;
+          strut->rect.y = 0;
+          break;
+        case META_SIDE_BOTTOM:
+          if (meta_screen_get_monitor_neighbor (screen, idx, META_SCREEN_DOWN))
+            continue;
+
+          strut->rect.height = screen->rect.height - strut->rect.y;
+          break;
+        case META_SIDE_LEFT:
+          if (meta_screen_get_monitor_neighbor (screen, idx, META_SCREEN_LEFT))
+            continue;
+
+          strut->rect.width += strut->rect.x;
+          strut->rect.x = 0;
+          break;
+        case META_SIDE_RIGHT:
+          if (meta_screen_get_monitor_neighbor (screen, idx, META_SCREEN_RIGHT))
+            continue;
+
+          strut->rect.width = screen->rect.width - strut->rect.x;
+          break;
+        }
+    }
+    
   /* Reordering doesn't actually matter, so we don't catch all
    * no-impact changes, but this is just a (possibly unnecessary
    * anyways) optimization */
@@ -1048,6 +1088,51 @@ meta_workspace_set_builtin_struts (MetaWorkspace *workspace,
   meta_workspace_invalidate_work_area (workspace);
 }
 
+void
+meta_workspace_update_snapped_windows (MetaWorkspace *workspace)
+{
+  GList *window_list = meta_workspace_list_windows (workspace);
+  GList *old = workspace->snapped_windows;
+  workspace->snapped_windows = NULL;
+
+  GList *iter;
+  MetaWindow *window;
+
+  for (iter = window_list; iter != NULL; iter = iter->next)
+  {
+    window = (MetaWindow *) iter->data;
+    if (window->tile_type == META_WINDOW_TILE_TYPE_SNAPPED)
+        workspace->snapped_windows = g_list_prepend (workspace->snapped_windows, window);
+  }
+  g_list_free (old);
+  g_list_free (window_list);
+
+  meta_workspace_recalc_for_snapped_windows (workspace);
+}
+
+gboolean
+meta_workspace_has_snapped_windows (MetaWorkspace *workspace)
+{
+    return g_list_length (workspace->snapped_windows) > 0;
+}
+
+void
+meta_workspace_recalc_for_snapped_windows (MetaWorkspace *workspace)
+{
+    GList *window_list = meta_workspace_list_windows (workspace);
+    GList *iter;
+    MetaWindow *win;
+    for (iter = window_list; iter != NULL; iter = iter->next)
+    {
+        win = META_WINDOW (iter->data);
+        if (meta_window_get_maximized (win))
+        {
+            meta_window_queue(win, META_QUEUE_MOVE_RESIZE);
+        }
+    }
+    g_list_free (window_list);
+}
+
 LOCAL_SYMBOL void
 meta_workspace_get_work_area_for_monitor (MetaWorkspace *workspace,
                                           int            which_monitor,
@@ -1109,6 +1194,8 @@ meta_motion_direction_to_string (MetaMotionDirection direction)
       return "Up-Left";
     case META_MOTION_DOWN_LEFT:
       return "Down-Left";
+    case META_MOTION_NOT_EXIST_YET:
+      return "Nothing";
     }
 
   return "Unknown";
@@ -1204,7 +1291,7 @@ meta_workspace_focus_default_window (MetaWorkspace *workspace,
     }
 
 
-  if (meta_prefs_get_focus_mode () == G_DESKTOP_FOCUS_MODE_CLICK ||
+  if (meta_prefs_get_focus_mode () == C_DESKTOP_FOCUS_MODE_CLICK ||
       !workspace->screen->display->mouse_mode)
     focus_ancestor_or_top_window (workspace, not_this_one, timestamp);
   else
@@ -1242,9 +1329,9 @@ meta_workspace_focus_default_window (MetaWorkspace *workspace,
                                                      window);
             }
         }
-      else if (meta_prefs_get_focus_mode () == G_DESKTOP_FOCUS_MODE_SLOPPY)
+      else if (meta_prefs_get_focus_mode () == C_DESKTOP_FOCUS_MODE_SLOPPY)
         focus_ancestor_or_top_window (workspace, not_this_one, timestamp);
-      else if (meta_prefs_get_focus_mode () == G_DESKTOP_FOCUS_MODE_MOUSE)
+      else if (meta_prefs_get_focus_mode () == C_DESKTOP_FOCUS_MODE_MOUSE)
         {
           meta_topic (META_DEBUG_FOCUS,
                       "Setting focus to no_focus_window, since no valid "
@@ -1299,7 +1386,7 @@ focus_ancestor_or_top_window (MetaWorkspace *workspace,
           meta_window_focus (ancestor, timestamp);
 
           /* Also raise the window if in click-to-focus */
-          if (meta_prefs_get_focus_mode () == G_DESKTOP_FOCUS_MODE_CLICK)
+          if (meta_prefs_get_focus_mode () == C_DESKTOP_FOCUS_MODE_CLICK)
             meta_window_raise (ancestor);
 
           return;
@@ -1308,7 +1395,7 @@ focus_ancestor_or_top_window (MetaWorkspace *workspace,
 
   window = meta_stack_get_default_focus_window (workspace->screen->stack,
                                                 workspace,
-                                                NULL);
+                                                not_this_one);
 
   if (window)
     {
@@ -1318,7 +1405,7 @@ focus_ancestor_or_top_window (MetaWorkspace *workspace,
       meta_window_focus (window, timestamp);
 
       /* Also raise the window if in click-to-focus */
-      if (meta_prefs_get_focus_mode () == G_DESKTOP_FOCUS_MODE_CLICK)
+      if (meta_prefs_get_focus_mode () == C_DESKTOP_FOCUS_MODE_CLICK)
         meta_window_raise (window);
     }
   else
@@ -1329,18 +1416,3 @@ focus_ancestor_or_top_window (MetaWorkspace *workspace,
                                               timestamp);
     }
 }
-
-/**
- * meta_workspace_get_screen:
- * @workspace: a #MetaWorkspace
- *
- * Gets the #MetaScreen that the workspace is part of.
- *
- * Return value: (transfer none): the #MetaScreen for the workspace
- */
-MetaScreen *
-meta_workspace_get_screen (MetaWorkspace *workspace)
-{
-  return workspace->screen;
-}
-
