@@ -31,6 +31,7 @@
 #include <meta/util.h>
 #include "core.h"
 #include "menu.h"
+#include "window-private.h"
 #include <meta/theme.h>
 #include <meta/prefs.h>
 #include "ui.h"
@@ -210,7 +211,7 @@ create_style_context (MetaFrames  *frames,
       provider = gtk_css_provider_get_named (theme_name, variant);
       gtk_style_context_add_provider (style,
                                       GTK_STYLE_PROVIDER (provider),
-                                      GTK_STYLE_PROVIDER_PRIORITY_THEME);
+                                      GTK_STYLE_PROVIDER_PRIORITY_SETTINGS);
     }
 
   g_free (theme_name);
@@ -1213,6 +1214,36 @@ meta_frame_titlebar_event (MetaUIFrame    *frame,
       }
       break;
 
+    case C_DESKTOP_TITLEBAR_ACTION_TOGGLE_STUCK:
+      {
+        meta_core_get (display, frame->xwindow,
+                       META_CORE_GET_FRAME_FLAGS, &flags,
+                       META_CORE_GET_END);
+
+        if (flags & META_FRAME_STUCK)
+          meta_core_unstick (display,
+                             frame->xwindow);
+        else
+          meta_core_stick (display,
+                           frame->xwindow);
+      }
+      break;
+
+    case C_DESKTOP_TITLEBAR_ACTION_TOGGLE_ABOVE:
+      {
+        meta_core_get (display, frame->xwindow,
+                       META_CORE_GET_FRAME_FLAGS, &flags,
+                       META_CORE_GET_END);
+
+        if (flags & META_FRAME_ABOVE)
+          meta_core_unmake_above (display,
+                                  frame->xwindow);
+        else
+          meta_core_make_above (display,
+                                frame->xwindow);
+      }
+      break;
+
     case C_DESKTOP_TITLEBAR_ACTION_MINIMIZE:
       {
         meta_core_get (display, frame->xwindow,
@@ -2582,10 +2613,14 @@ get_control (MetaFrames *frames,
   MetaFrameGeometry fgeom;
   MetaFrameFlags flags;
   MetaFrameType type;
+  MetaWindow *window;
   gboolean has_vert, has_horiz, has_left, has_right, has_bottom, has_top;
   gboolean has_north_resize;
   cairo_rectangle_int_t client;
 
+  window = meta_core_get_window (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
+                                 frame->xwindow);
+
   meta_frames_calc_geometry (frames, frame, &fgeom);
   get_client_rect (&fgeom, fgeom.width, fgeom.height, &client);
 
@@ -2625,6 +2660,11 @@ get_control (MetaFrames *frames,
 
   if (POINT_IN_RECT (x, y, fgeom.max_rect.clickable))
     {
+      if (flags & META_FRAME_MAXIMIZED &&
+          (META_WINDOW_TILED_TOP (window) ||
+           META_WINDOW_TILED_BOTTOM (window)))
+        return META_FRAME_CONTROL_MAXIMIZE;
+
       if (flags & META_FRAME_MAXIMIZED)
         return META_FRAME_CONTROL_UNMAXIMIZE;
       else
